/*
 * Screen editor
 */

#include <stdio.h>
#include "../sedit.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#ifdef linux
#include <termio.h>
#else
#include <termios.h>
#endif
#include <signal.h>

#define VERSION	0x15

#ifndef isdigit
#define isdigit(c) ((c)>='0' && (c)<='9')
#endif

unsigned char *mem_buf;				/* Start of text */
unsigned char *end_buf;				/* Points to null after last line */
int top_lin;				/* Line number for line at top of screen */
int bot_lin;				/* Number of line beyond last displayed */
int cur_lin;				/* Current */
unsigned char *top_lp;				/* Points to source for top line */
unsigned char *bot_lp;				/* Points to the string beyond the last displayed */
unsigned char *cur_lp;				/* Start of current line */
unsigned char *cur_ptr;				/* Current working pointer */
int cur_off;				/* cur_ptr - cur_lp (column offset of cursor) */
int save_coff;				/* Flag for preserving cur_off */
int bot_y;					/* Display line at which no text is displayed */
int top_y;					/* Display line of top displayed line */
int cur_y;					/* Current Y cursor pos */
int cur_x;					/* Current X cursor pos */
int lin_siz;				/* Number of screen lines occupied by a string */
int lin_dlen;				/* Number of screen unsigned chars occupied by a string */
int scrnleft;				/* Used for preventing printing from going off the bottom */
int max_mem;				/* Maximum buffer size */
int lin_len;				/* Number of real unsigned chars in a string */
int width;					/* Screen width */
int height;					/* Screen height */
int ibindex;				/* Index for insert buffer */
int nfound;					/* Number of successful searches/substitutes */
int dlscren;				/* Flag: Enable scrolling in dlscren() */
int modflag;				/* Flag: file modified */
int wasqp;					/* Flag: last command was ^QP */
unsigned char *blkstart;			/* Start of block */
unsigned char *blkend;				/* End of block */
unsigned char *newcur_ptr;			/* Used for keeping track of where we are */
unsigned char **markers;			/* Markers */
int blkset;					/* Flag: block is marked */
int hilit;					/* Flag: highlight mode is set */
int scr_top;				/* Current screen top */
int frozen;					/* Flag: screen frozen */
int quitflag;				/* Flag: time to exit */
int tabwidth;				/* Optional tab size */
int optNoviceMode;			/* Novice mode */
char *newfilename;			/* Name of file to switch to */
unsigned char *stringtolookfor = 0;	/* If non-zero, look for this string on start of file */
int look_for_line = 0;
int line_to_look_for = 0;
unsigned char *tagtolookfor = 0;	/* Set by -t option */
int nodrawonsearch = 0;		/* If set, search() does not redraw */
int needgotsigwinch;
int auto_format;
int start_at_top;

extern int dispcontrol;
extern int overstrike;		/* Characters overtype current ones */
extern unsigned lmargin, rmargin;
extern int curscrtop;
extern int ismgr;
extern int inleditor;
extern int sigwinchkludge;

extern struct sr_data search_sr_data, subs_sr_data;

/* Flags for attributes */
int attrnow, attrwant;
int tabwgiven, nondocmode = 1;

unsigned char *file_name;
unsigned char *statmes;				/* Top left status message */

extern int justifymode;

unsigned char *getnext(), *get_prev();
unsigned char *fwd_fnw(), *rev_fnw();
extern char *roomfor();
extern char *getenv();
extern char *patchenv();

unsigned char *_markers[10];		/* Markers */
unsigned char _file_name[SEDLINESIZE];
unsigned char prevfilename[SEDLINESIZE];
unsigned char _statmes[100];		/* Top left status message */
struct stat statent;

char *ourname;
int autotabs;
int docmodegiven;
long efilestat;				/* Status bits of edited file */
int euserid;				/* User ID of edited file */

int lowbaud;				/* True if running at low baud rates */
int statdelay;				/* Ticks to delay before drawing status line */

extern int beingEdited;

static char *cutFileName = 0;
static char **fnames = 0;

static char invertregexp = 0;
static char usuallywrap = 0;

int unlink_input = 0;

int
invertRegexp()
{
	return invertregexp;
}

int usuallyWrap()
{
	return usuallywrap;
}

static char *noInputFile()
{
	char *newfilename;

	if (isatty(fileno(stdin)))
	{
		addRollFile(newfilename = roomfor("NoName"));
	}
	else
	{
		char temp_file[32];
		int ok = 0;

		sprintf(temp_file, "dde%d", getpid());

		{
			FILE *f = fopen(temp_file, "w+");
			if (f)
			{
				int c;

				while ((c = fgetc(stdin)) != EOF)
					fputc(c, f);
				fclose(f);
				fclose(stdin);
				fclose(stdout);
				fopen("/dev/tty", "r");
				fopen("/dev/tty", "a");
				ok = 1;
			}
		}

		if (ok)
		{
			addRollFile(newfilename = roomfor(temp_file));
			unlink_input = 1;
		}
		else
		{
			exit(1);
		}
	}
	return newfilename;
}

/*
 * Main entry point
 */

main(argc, argv)
int argc;
char **argv;
{
	register int fsize, retval, addr;
	int realfsize;
	int arg, val;
	register char *cp, ch;
#ifdef __STDC__
	void sighndl(int), sigwinch(int);
#else
	void sighndl(), sigwinch();
#endif
	int nfnames = 0;
	char badopt = 0;

	ourname = argv[0];

	captureinit();
	ddercInit();
#if 0
	termstart();
#endif

	{
		char *home = getenv("HOME");
		if (home == 0)
			home = ".";
		if (cutFileName)
			Free(cutFileName);
		cutFileName = Malloc(strlen(home) + 20);
		sprintf(cutFileName, "%s/.ddeCutBuffer", home);
	}

	markers = _markers;
	statmes = _statmes;

	file_name = _file_name;
	*file_name = '\0';

#if 0
	sedinit();
#endif

	{
		char *tw = getenv("TABWIDTH");

		tabwidth = tw ? atoi(tw) : 5;
		if (tw == 0)
			tabwidth = 5;
		else
			tabwgiven = 1;
	}

	autotabs = 0;
	docmodegiven = 0;
	fnames = Malloc(argc * sizeof(*fnames));

	lowbaud = 0;
	statdelay = 20;				/* Default */

	for (arg = 1; arg < argc; arg++)
	{
		cp = argv[arg];
		if (*cp == '-' || *cp == '+')
		{
			if (*cp != '+')
				cp++;
			switch (ch = *cp++)
			{
			case 'A':	/* Set auto format */
				auto_format++;
				break;
			case 'a':	/* Set autotabs */
				autotabs = 1;
				break;
			case 'd':	/* Document mode */
				nondocmode = 0;
				docmodegiven = 1;
				break;
			case 'i':	/* Overstrike mode */
				overstrike = 1;
				break;
			case 'j':	/* Set justify mode */
				if (!strcmp(cp, "rr"))
					justifymode = J_RAGGED;
				else if (!strcmp(cp, "rj"))
					justifymode = J_RIGHT;
				else if (!strcmp(cp, "cj"))
					justifymode = J_CENTRE;
				else usage(1);
				break;
			case 'N':
				optNoviceMode++;
				break;
			case 'F':	/* tag file */
			case 'h':	/* Screen height */
			case 'l':	/* Set left margin */
			case 'r':	/* Set right margin */
			case 'w':	/* Width */
			case 'T':	/* Set tabwidth */
			case 'P':	/* Push a string into input */
			case '+':	/* Search / line number */
				if (!*cp)
				{
					cp = argv[++arg];
					if (arg == argc)
						usage(2);
				}
				val = atoi(cp);
				switch (ch)
				{
				case 'F':
					addTagFile(cp);
					break;
				case 'h':
					height = val;
					break;
				case 'l':
					lmargin = val - 1;
					break;
				case 'r':	/* Set right margin */
					rmargin = (unsigned)val - 1;
					break;
				case 'w':
					width = val;
					break;
				case 'T':	/* Set tabwidth */
					tabwidth = val;
					tabwgiven = 1;
					break;
				case 'P':
					stringtome(cp);
					break;
				case '+':
					sendtome(CTL('Q'));
					if (isdigit(*cp))
					{	/* Go to line number */
						sendtome('g');
						stringtome(cp);
						sendtome('\r');
					}
					else
					{	/* Search */
						char *ecp;

						if (*cp == '/')
							cp++;
						ecp = cp;
						while (*ecp && ecp[1])
							ecp++;
						if (*ecp == '/')
							*ecp = '\0';
						sendtome('F');	/* Make it a regexp search */
						stringtome(cp);
						stringtome("\r\r");
					}
					break;
				}
				break;
			case 'n':	/* Non-document mode */
				nondocmode = 1;
				docmodegiven = 1;
				if (!tabwgiven)
					tabwidth = 8;
				break;
			case 's':	/* No status line */
				curscrtop = 0;
				scr_top = 0;
				break;
			case 't':	/* Startup tag */
				if (!*cp)
				{
					cp = argv[++arg];
					if (arg == argc)
						usage(3);
				}
				tagtolookfor = (unsigned char *)cp;
				break;
			case 'q':
				usage(99);
				break;
			case 'R':
				invertregexp = 1;
				break;
			case 'S':
				start_at_top++;
				break;
			case 'W':
				usuallywrap = 1;
				break;
			case '?':
				usage(8);
				break;
			default:
				fprintf(stderr, "%s: %s: Bad option\n",
					ourname, argv[arg]);
				badopt = 1;
				break;
			}
		}
		else
		{
			nfnames++;
			addRollFile(cp);
		}
	}

	if (nfnames == 0 && !tagtolookfor)
	{
		newfilename = noInputFile();
	}
	else if (nfnames)
	{
		newfilename = roomfor(firstRollFile());
	}

	termstart();
	sedinit();

	if (badopt)
	{
		fflush(stderr);
		sleep(1);
	}

	if (!tabwidth)
		tabwidth++;

	/* Do signals here */
	signal(SIGINT, sighndl);
	signal(SIGQUIT, sighndl);
	signal(SIGWINCH, sigwinch);

	while (newfilename != (char *)0 || (!nfnames && tagtolookfor))
	{
		char *efname = 0, *_efname;
		char efset = 0;		/* Flag: must free efname */

		if (newfilename)
		{
			efname = roomfor(newfilename);
			efset = 1;
			Free(newfilename);
			newfilename = 0;
		}
		else
		{	/* (!nfnames && tagtolookfor):  Look for the tag */
				unsigned char *name, *search;
				int is_line_number;
				if (lookuptag(tagtolookfor, &name, &search, &look_for_line))
				{
					efname = (char *)name;
					efset = 0;
					if (look_for_line)
					{
						line_to_look_for = atoi(search);
					}
					else
					{
						stringtolookfor = search;
					}
				}
				else
				{
					efname = "Unknown";
					efset = 0;
				}
				tagtolookfor = 0;
		}

		_efname = patchenv(efname);
		if (efset)
		{
			Free(efname);
			efset = 0;
		}
		efname = _efname;
		efset = 1;

		if ((addr = stat(efname, &statent)) < 0)
		{
			fsize = 0;
			efilestat = -1;
			euserid = -1;
		}
		else if (S_ISDIR(statent.st_mode))
		{
			fprintf(stderr, "%s is a directory\n", efname);
			sleep(1);
			fsize = 0;
			efilestat = -1;
			euserid = -1;
			if (efset)
				Free(efname);
			efname = "Unknown";
			efset = 0;
		}
		else
		{
			fsize = statent.st_size;
			efilestat = statent.st_mode;
			euserid = statent.st_uid;
		}
		realfsize = fsize;
		fsize = (fsize + 0x80000) & ~0x3ff;
		mem_buf = Malloc(fsize);
		if (mem_buf == 0)
		{
			fprintf(stderr, "Insufficient memory\n");
			doexit(-1);
		}
		retval = tosedit(efname, fsize, realfsize);
		/* Salt marked block away in undo buffers */
		if (blkok())
			predelete(blkstart, blkend - blkstart);
		else
			sstatmes("");

		Free(mem_buf);
		if (efset)
		{
			Free(efname);
			efset = 0;
		}
		flushop();
	}
	termterm();
	doexit(retval);
}

int length_to_height(len)
{
	return (len - (term_scrolls_at_last_col() ? 0 : 1)) / width;
}

unsigned char ibuf[IBSIZE+50];	/* Insert buffer */
static char endcapturechars[] = {
	CTL(']'), CTL('L'), CTL('N'), CTL('U'), CTL('E'), CTL('R'), CTL('X'),
	CTL('C'), CTL('W'), CTL('Z'), 033, CTL('K'), CTL('Q'), CTL('O'), CTL('Y'), 0 };
static char startcapturechars[] = {
	CTL('G'), CTL('H'), '\t', '\r', CTL('P'), CTL('T'), CTL('V'), 0x7f, 0 };

tosedit(name, edsize, realfsize)
char *name;
int edsize;
{
	register int i;
	register unsigned char ch, nxtch;
	register unsigned char *tempptr;
	register unsigned char isdel;
	unsigned char insch = 0;		/* Flag: unsigned char was inserted */
	int oldlin_dlen;				/* Previous displayed line length */
	int nexty;						/* Where line after current starts */
	unsigned char *lastcptr;		/* Last setting of cur_ptr */
	int distance;
	char capturing = 0;
	int leave_it_alone = 0;

	clr_scrn();
	clearmarkers();
	quitflag = 0;
	max_mem = edsize;
	end_buf = mem_buf;
	cur_lp = mem_buf;
	/* Clear the ^QP stack */
	qpstackinit();
	if (stringtolookfor || look_for_line)
	{
		cur_ptr = mem_buf;
	}
	else
	{
		cur_ptr = mem_buf + findOldFile(name);
		if (cur_ptr >= mem_buf + realfsize)
		{
			cur_ptr = end_buf - 2;
			if (cur_ptr < mem_buf)
				cur_ptr = mem_buf;
		}
	}

	movstr(name, file_name, SEDLINESIZE);
	redrawmenubar();	/* MGR special: we have the filename now */
	file_name[SEDLINESIZE-1] = '\0';
	if (read_file(file_name, 0, 1) < 0)
		blkswap(cur_ptr, 0, "\n", 1);
	if (!docmodegiven && strchr(mem_buf, '\t'))
	{
		nondocmode = 1;
		if (!tabwgiven)
			tabwidth = 8;
	}
	setpointers(name);

	if (start_at_top)
	{
		start_at_top = 0;
		cur_ptr = mem_buf;
	}

	if (beingEdited)
	{
		sstatmes("\007File being edited");
		beingEdited = 0;
	}

	if (stringtolookfor)
	{
		cur_ptr = mem_buf;
		strncpy(search_sr_data.old, stringtolookfor, SEDLINESIZE);
		search_sr_data.new[SEDLINESIZE - 1] = '\0';
		search_sr_data.useregexp = 0;
		search_sr_data.reverse = 0;
		search_sr_data.ignorecase = 0;
		search_sr_data.fromtop = 0;
		search_sr_data.wrap = 0;
		search_sr_data.ntimes = 1;
		search_sr_data.initialised = 1;
		nodrawonsearch = 1;
		search(&search_sr_data);
		nodrawonsearch = 0;
		Free(stringtolookfor);
		stringtolookfor = (unsigned char *)0;
	}
	if (cur_ptr < mem_buf || cur_ptr >= end_buf)
		cur_ptr = mem_buf;

	/*
	 * Look for special mode definition sequences
	 * on the first line of the file
	 */
	searchtw();
	/* Make things sensible */
	if (tabwidth < 1 || tabwidth > 99)
		tabwidth = 8;
	if (lmargin == 0 && rmargin == 0)
	{
		lmargin = 0;
		rmargin = 60;
	}

	if (lmargin > width - 5)	/* Unsigned! */
		lmargin = width - 5;
	if ((rmargin <= lmargin + 1) || (rmargin > width - 1))
		rmargin = width - 1;

	if (nondocmode)
		dispcontrol = 1;
	lastcptr = cur_ptr;
	modflag = ibindex = nxtch = save_coff = 0;
	redoptr();
	redwhole();
	if (look_for_line)
	{
		gotoLine(line_to_look_for);
		look_for_line = 0;
	}

	for ( ; ; )
	{
		static int oldmodflag = 0;

		if (quitflag)
		{
			saveFileState(file_name);
			return 0;	/* The editor exit point */
		}

		if (ibindex == 0)	/* No chars to be inserted */
		{
			set_param(cur_lp);
			oldlin_dlen = lin_dlen;
			if (!save_coff)
				cur_off = cur_x;
			isdel = save_coff = 0;
		}

		if (leave_it_alone)
		{
			leave_it_alone = 0;
			oldmodflag = modflag;	/* Stops \r making next motion reformat */
		}
		else if (nondocmode && auto_format)
		{
			unsigned char *p;

			mergeinsbuffer();
			if (modflag && modflag != oldmodflag)
			{
				p = cur_lp;
				while (p < end_buf && *p != '\n')
					p++;

				if (p - cur_lp > rmargin)
				{	/* Overflowed the line */
	/*				fprintf(stderr, "overflowed\n");*/
					nondocreformat(1);
				}
				else if (*p == '\n' && p > mem_buf - 1 && p[-1] != '\n')
				{	/* Underflowed? */
					int col = p - cur_lp;
	/*				fprintf(stderr, "init col=%d\n", col);*/
					p++;
					while (p < end_buf && *p == ' ')
						p++;
					if (*p != '\n')
					{
						col++;
	/*
						fprintf(stderr,
							"after allowing for blank, col=%d\n", col);
	*/
						while (p < end_buf && !blank(*p))
						{
							p++;
							col++;
						}
	/*					fprintf(stderr, "after word skip, col=%d\n", col);*/
						if (p < end_buf && col <= rmargin)
						{
	/*						fprintf(stderr, "reformatting\n");*/
							nondocreformat(1);
						}
					}
				}
			}
			oldmodflag = modflag;
		}

		/* How far has the cursor moved since last time? */
		distance = cur_ptr - lastcptr;
		if (distance < 0)
			distance = -distance;
		if (distance > 100 && !wasqp)		/* Far enough to move the mark? */
			addqpstack(lastcptr);
		wasqp = 0;
		lastcptr = cur_ptr;

		if (attrnow && !nondocmode)
		{
			attrwant = 0;	/* Clear all attributes */
			setbul();
		}
		statline(cur_x, cur_y);

		*statmes = '\0';
		nexty =	cur_y + length_to_height(oldlin_dlen) + 1;
		gotoxy(cur_x, cur_y);
		if (!nondocmode)
			setattr();		/* Work out bold state, etc */
again:
		ch = (nxtch) ? nxtch : rawgetchar();

		if (ch & 0x80)
		{
			int leaveit = 0;

			switch (ch)
			{
			case ALT(CTL('D')):
				sendtome(CTL('D'));
				break;
			case ALT('5'):
				sendtome(CTL('Q'));
				sendtome('%');
				break;
			case ALT('9'):			/* Go back a block */
				sendtome(CTL('O'));
				sendtome('(');		/* ) */
				break;
			case ALT('0'):			/* Go fwd a block */
				sendtome(CTL('O'));
/* ( */			sendtome(')');
				break;
			case ALT('['):			/* ] */
			case ALT('{'):			/* } */
				sendtome(CTL('Q'));
				sendtome('[');		/* ] */
				break;
/* [ */		case ALT(']'):
/* { */		case ALT('}'):
				sendtome(CTL('Q'));
/* [ */			sendtome(']');
				break;
			case ALT('='):
				sendtome(CTL('K'));
				sendtome('=');
				break;
			case ALT('b'):
				sendtome(CTL('O'));
				sendtome(CTL('b'));
				break;
			case ALT('c'):
				if (blkok())
				{
					sendtome(CTL('K'));
					sendtome(CTL('W'));
					stringtome(cutFileName);
					sendtome('\r');
					sendtome(CTL('K'));
					sendtome('h');
				}
				else
					statline(cur_x, cur_y);
				break;
			case ALT('f'):
				sendtome(CTL('O'));
				sendtome(CTL('f'));
				break;
			case ALT('h'):
				sendtome(CTL('K'));
				sendtome('h');
				break;
			case ALT('o'):
				sendtome(CTL('q'));
				sendtome('d');
				sendtome('\n');
				break;
			case ALT('O'):
				sendtome(CTL('q'));
				sendtome('s');
				sendtome('\n');
				sendtome(CTL('s'));
				break;
			case ALT('p'):		/* Insert current filename */
				stringtome(currentFileName());
				break;
			case ALT('P'):		/* Insert current pathname */
				stringtome(currentPathName());
				break;
			case ALT('q'):
				if (modflag || insch)
				{
					sendtome(CTL('K'));
					sendtome(CTL('D'));
					sendtome('\r');
				}
				sendtome(CTL('K'));
				sendtome(CTL('Q'));
				break;
			case ALT('s'):
				if (modflag || insch)
				{
					sendtome(CTL('K'));
					sendtome(CTL('D'));
					sendtome('\r');
				}
				break;
			case ALT('v'):
				sendtome(CTL('K'));
				sendtome(CTL('R'));
				stringtome(cutFileName);
				sendtome('\r');
				break;
			case ALT('w'):
				sendtome(CTL('W'));
				sendtome(CTL('E'));
				break;
			case ALT('x'):
				if (blkok())
				{
					sendtome(CTL('K'));
					sendtome(CTL('W'));
					stringtome(cutFileName);
					sendtome('\r');
					sendtome(CTL('K'));
					sendtome(CTL('Y'));
				}
				else
					statline(cur_x, cur_y);
				break;
			case ALT('z'):
				sendtome(CTL('Z'));
				sendtome(CTL('X'));
				break;
			default:
				leaveit = 1;
			}
			if (leaveit == 0)
				goto again;
		}

		/*
		 * An insert/delete starts capturing.  A larg(ish)
		 * cursor movement stops it
		 */
		if (!capturing)
		{
			if (ch >= ' ' || strchr(startcapturechars, ch))
			{
				capture(ch);
				capturing = 1;
			}
		}
		else
		{	/* Capturing */
			if (strchr(endcapturechars, ch))
			{
				stopcapture();
				capturing = 0;
			}
			else
			{
				capture(ch);
			}
		}

		nfound = nxtch = insch = 0;
		if (ch == '\n')
			ch = '\r';
		if (ch == CTL(' '))
			ch = ' ';

		if (ch == '\r' && auto_format)
		{
			leave_it_alone = 1;
		}

		if (ch == '\r')
		{
			set_param(cur_lp);
			nexty = cur_y + lin_siz;
		}

		if (	((ch >= ' ') && (ch != 0x7f) && ((ch & 0x80) == 0)) ||
				(ch == '\r') ||
				(ch == '\t') ||
				(ch == 0x10))
		{	/* An insertable character */
			sethil(cur_ptr);	/* Set highlight mode if necessary */
			if (ch == CTL('P'))
			{
				sstatmes("^P");
				statline(cur_x, cur_y);
				ch = rawgetchar();
				statmes[0] = '\0';
				statline(cur_x, cur_y);
			}
			if (overstrike && (ch != '\r'))
			{
				if ((ch >= ' ') && (MASK(*cur_ptr) >= ' '))
				{	/* Simple overstrike */
					modflag++;
					*cur_ptr = ch;
					opchar(ch);
					opchar('\b');
					ch = CTL('D');	/* Revolting */
					goto cntld;
				}
				if (MASK(*cur_ptr) != '\n')
				{
					delete(1);
					set_param(cur_lp);
					gotoxy(cur_x, cur_y);
				}
			}
			insch = 1;		/* A char has been added */
			if (nondocmode || (ch != '\t'))
				ibuf[ibindex++] = (ch == '\r') ? '\n' : ch;
			if (ch != '\r')
			{
				if (ch == '\t')
				{
					if (nondocmode)
					{
						do
							opchar(' ');
						while ((++cur_x) % tabwidth);
					}
					else
					{
						int offs, ibi;
						unsigned char *cp;
						cp = cur_lp;
						offs = 0;
						while (cp < cur_ptr)
						{
							if (*cp++ >= ' ')
								offs++;
						}
						for (ibi = 0; ibi < ibindex; ibi++)
						{
							if (ibuf[ibi] >= ' ')
								offs++;
						}
						do
						{
							opchar(' ');
							ibuf[ibindex++] = ' ';
							++cur_x;
						}
						while ((++offs) % tabwidth);
					}
				}
				else
				{
					if (ch < ' ')
					{
						if (dispcontrol)
						{
							opchar('^');
							opchar(ch + 0x40);
							cur_x += 2;
						}
						else if (ch == CC_BOLD)
						{
							if (attrnow & BOLD)
								attrwant &= ~BOLD;
							else
								attrwant |= BOLD;
							setbul();
						}
						else if (ch == CC_UNDERLINE)
						{
							if (attrnow & UNDERLINE)
								attrwant &= ~UNDERLINE;
							else
								attrwant |= UNDERLINE;
							setbul();
						}
						else if (ch == CC_SUBSCRIPT)
						{
							if (attrnow & SUBSCRIPT)
								attrwant &= ~SUBSCRIPT;
							else
								attrwant |= SUBSCRIPT;
							setbul();
						}
						else if (ch == CC_SUPERSCRIPT)
						{
							if (attrnow & SUPERSCRIPT)
								attrwant &= ~SUPERSCRIPT;
							else
								attrwant |= SUPERSCRIPT;
							setbul();
						}
						else if (ch == CC_ITALICS)
						{
							if (attrnow & ITALICS)
								attrwant &= ~ITALICS;
							else
								attrwant |= ITALICS;
							setbul();
						}
						else if (ch == CC_WIDE)
						{
							if (attrnow & WIDE)
								attrwant &= ~WIDE;
							else
								attrwant |= WIDE;
							setbul();
						}
					}
					else
					{
						opchar(ch);
						cur_x++;
					}
				}
				/* New length of this line */
				dlscren = 1;	/* Enable scrolling by draw_lin */
				lin_dlen = draw_lin(cur_ptr, cur_x, cur_y);
				dlscren = 0;

				if (length_to_height(lin_dlen) > length_to_height(oldlin_dlen))
				{	/* New line created */
					oldlin_dlen = lin_dlen;
					if ((cur_y + length_to_height(lin_dlen) + 1) > height)
					{	/* draw_lin() caused a scroll */
						cur_y--;
						if (--top_y < scr_top)
						{	/* Top line pushed off */
							set_param(top_lp);
							top_y += lin_siz;
							top_lp = getnext(top_lp);
							top_lin++;
						}
					}
					else
					{	/* Use insert line here one day */
						draw(	getnext(cur_lp),
								cur_lin+1,
								cur_y + length_to_height(oldlin_dlen) + 1);
					}
				}
				else
					oldlin_dlen = lin_dlen;
			}	/* ch == '\r' */
			cur_off = cur_x;
		}	/* ch insertable */

cntld:

	/*
	 * If ch was a cr or ch non-insertable or buffer full, move
	 * the insert buffer in
	 */
		if ((ch == '\r') || (!insch && ibindex) || (ibindex >= IBSIZE))
		{	/* Time to move the insert buffer in */
			mergeinsbuffer();
			if (ch == '\r')
			{
				int new_y_size;
				if (nondocmode && autotabs)
				{	/* Calculate tab count for autoindent */
					unsigned char *p;
					p = cur_lp;
					while (*p++ == '\t')
						sendtome('\t');
					if (cur_ptr > mem_buf + 2 && cur_ptr[-2] == '{')	/* } */
						sendtome('\t');
				}
				/* CR inserted */
				clear_eol();
				/* Produce the '<' in rightmost column */
				if (!nondocmode &&
					(cur_y + length_to_height(cur_x)) < height - 1)
				{
					gotoxy(	width - 1,
							cur_y + length_to_height(cur_x));
					opchar('<');
					gotoxy(cur_x, cur_y);
				}
				set_param(cur_lp);
				cur_y += lin_siz;	/* Move onto newly created line */
				cur_lp = getnext(cur_lp);
				if ((++cur_lin >= bot_lin++))
				{	/*zz This line is the last one */
					bot_lin = cur_lin;
					bot_lp = cur_lp;
					bot_y = cur_y;
					scrl_up();
				}
				else
				{	/* There are lines below this one */
					set_param(cur_lp);
					if ((cur_y + lin_siz) > nexty)
					{	/* It is necessary to move lines down */
						gotoxy(0, nexty);
						inslin(nexty);
						if (++bot_y >= height)
						{
							set_param(bot_lp = get_prev(bot_lp));
							bot_lin--;
							bot_y -= lin_siz;
						}
					}
					draw_lin(cur_lp, 0, cur_y);
				}
			}
			if (!insch) set_param(cur_lp);
		}

		if (insch)
			goto skipsw;

		switch (ch)
		{
		case 0:		/* Null char */
			break;
		case CTL(']'):		/* Use char under cursor as tag */
			gototag(0);
			break;
		case ALT('t'):
			gototag((char *)1);
			break;
		case CTL('L'):
			search_sr_data.ntimes = 1;
			search(&search_sr_data);
			break;
		case CTL('N'):
			subs_sr_data.ntimes = 1;
			substitute(&subs_sr_data);
			break;
		case CTL('H'):
		case 0x7f:
			isdel = 1;
		case CTL('S'):
			--cur_ptr;
			goto rev;
		case CTL('A'):
			cur_ptr = rev_fnw(cur_ptr);
	rev:	if (cur_ptr < mem_buf)
			{
				cur_ptr = mem_buf;
				if (isdel)	/* At top of file */
					goto revout;
			}
			if (cur_ptr < cur_lp)
			{	/* Backed up a line */
				cur_lin--;
				cur_lp = get_prev(cur_lp);
				set_param(cur_lp);
				cur_y -= lin_siz;
				if (cur_lp < top_lp)
					scrl_dn();
				set_param(cur_lp);	/* For delete() usage */
			}
			if (isdel)
				delete(1);	/* Handle DEL */
revout:		break;
		case CTL('U'):
			sstatmes("^U");
			statline(cur_x, cur_y);
			ch = rawgetchar();
			if (ch == CTL('U'))		/* Double ^U */
				ch = 0;
			else
				ch -= '0';
			if (putundobuf(ch) < 0)
				sstatmes("\007Can't put buffer");
			break;
		case CTL('V'):
			if (cur_ptr > cur_lp)
			{
				i = cur_ptr - cur_lp;	/* No. of unsigned chars */
				cur_ptr = cur_lp;
				delete(i);
			}
			break;
		case CTL('D')|0x80:
		case CTL('D'):
			if (cur_ptr == end_buf)
				break;
			cur_ptr++;
			goto fwd;
		case CTL('F'):
			if (cur_ptr == end_buf)
				break;
			cur_ptr = fwd_fnw(cur_ptr);
fwd:		if (cur_ptr == end_buf)
			{
				cur_ptr -= 1;
				break;
			}
			if	(
					(cur_ptr >= (tempptr = getnext(cur_lp))) &&
					!((cur_ptr == end_buf) && !endatcr())
				)
			{	/* Have moved to another line */
				cur_lp = tempptr;
				cur_lin++;
				cur_y += lin_siz;
				if (cur_lp == bot_lp)
					scrl_up();
				set_param(cur_lp);
			}
			break;
		case CTL('E'):
			cntrle();
			break;
		case CTL('R'):
			nxtch = i = 0;
			while ((i < mpaglen()) && (cur_lp != mem_buf))
			{
				cur_ptr = cur_lp = get_prev(cur_lp);
				set_param(cur_lp);
				cur_y -= lin_siz;
				cur_lin--;
				i += lin_siz;
			}
			fix_cpos();
			redwhole();
			break;
		case CTL('X'):
			cntrlx(1);
			break;
		case CTL('C'):
			nxtch = i = 0;
			while ((i < mpaglen()) &&
					((tempptr = getnext(cur_ptr)) != end_buf))
			{
				cur_ptr = cur_lp = tempptr;
				cur_lin++;
				cur_y += lin_siz;
				set_param(cur_lp);
				i += lin_siz;
			}
			fix_cpos();
			redwhole();
			break;
		case CTL('W'):
			if (cur_lin == (bot_lin-1))
				cntrle();
			else
				fix_cpos();
			if (top_lin != 1)
			{
				dn_scrl();
				top_y++;
				cur_y++;
				bot_y++;
				if (bot_y > height)
				{	/* Bottom line pushed off */
					bot_lin--;
					set_param(bot_lp = get_prev(bot_lp));
					bot_y -= lin_siz;
				}
				set_param(tempptr = get_prev(top_lp));
				if (lin_siz <= (top_y - scr_top))
				{	/* Can fit another line */
					top_y -= lin_siz;
					top_lin--;
					top_lp = tempptr;
					draw_lin(top_lp, 0, top_y);
				}
			}
			break;
		case CTL('Z'):
			if (cur_lin == top_lin)
				cntrlx(1);
			else
				fix_cpos();
			if (bot_lp != end_buf)
			{
				up_scrl(1);
				top_y--;
				cur_y--;
				bot_y--;
				if (top_y < scr_top)
				{	/* Top line pushed off */
					set_param(top_lp);
					top_y += lin_siz;
					top_lin++;
					top_lp = getnext(top_lp);
				}
				draw_lin(bot_lp, 0, bot_y);
				set_param(bot_lp);
				if ((bot_y + lin_siz) <= height)
				{
					bot_y += lin_siz;
					bot_lin++;
					bot_lp = getnext(bot_lp);
				}
			}
			break;
		case 0x1b:	/* ESC */
			doEscape();
			break;
		case CTL('K'):
			cntrlkq('K');
			break;
		case CTL('Q'):
			cntrlkq('Q');
			break;
		case CTL('O'):
			cntrlkq('O');
			break;
		case CTL('\\'):
			overstrike = !overstrike;
			beep();
			break;
		case CTL('B'):
			tempptr = getnext(cur_ptr);
			if (cur_ptr == cur_lp)
				cur_ptr = tempptr - 1;
			else	/* Go to start */
				cur_ptr = cur_lp;
			break;
		case CTL('G'):
			delete(1);
			break;
		case CTL('T'):
			delete(fwd_fnw(cur_ptr) - cur_ptr);
			break;
		case CTL('Y'):
			tempptr = getnext(cur_lp);
			if (tempptr == end_buf)
			{	/* Last line */
				cur_ptr = cur_lp;
				delete(tempptr - cur_lp - 1);
			}
			else
			{
				i = tempptr - cur_lp;
				predelete(cur_lp, i);
				blkswap(cur_lp, i, cur_lp, 0);
				cur_ptr = cur_lp;
				dellines(cur_y, lin_siz);
				bot_y -= lin_siz;
				bot_lp -= i;
				bot_lin--;
				if (cur_lin == bot_lin)
					scrl_up();
				else
				{
					set_param(tempptr = get_prev(bot_lp));
					draw(tempptr, bot_lin - 1, bot_y - lin_siz);
				}
			}
			break;
		default:
			bad_cmd();
			break;
		}
skipsw:
		continue;
	}
}

doEscape()
{
/*	eschit(); */
	unfreeze(0);
	redwhole();
}

mergeinsbuffer()
{
	if (ibindex)
	{
		blkswap(cur_ptr, 0, ibuf, ibindex);
		cur_ptr += ibindex;
		bot_lp += ibindex;
		ibindex = 0;
	}
}

int _signalled;
void sighndl(sig)
int sig;
{
	signal(sig, sighndl);
	_signalled = 1;
}

/* Added a comment */

gotsigwinch()
{
	eschit();
	statline(cur_x, cur_y);
	flushop();
	needgotsigwinch = 0;
}

maybegotsigwinch()
{
	if (needgotsigwinch)
		gotsigwinch();
}

void sigwinch(sig)
int sig;
{
	if (inleditor)
		needgotsigwinch = 1;
	else
		gotsigwinch();

	sigwinchkludge = 1;

	signal(SIGWINCH, sigwinch);
}

signalled()
{
	int ret;

	ret = _signalled;
	_signalled = 0;
	return ret;
}

usage(diag)
{
	if (diag != 0)
		termterm();

fprintf(stderr,"\n1616 word processor V%d.%d. (%d)\n",
				VERSION / 16, VERSION & 15, diag);

fprintf(stderr,"Copyright (c) 1989, 1990 Andrew Morton, Applix pty limited\n");

fprintf(stderr, "\n");
fprintf(stderr, "Usage: %s [-AadinqRsW] [+NN] [+/pat/] [-F tagfile] [-T tabwidth] [-t tag]\n", ourname);
fprintf(stderr, "              [-jrr] [-jrj] [-jcj] [-l NN] [-r NN] [-w NN]\n");
fprintf(stderr, "              [-h NN] filename [filename ...]\n");

fprintf(stderr,"\t-A     :\tEnable autoformatting\n");
fprintf(stderr,"\t-a     :\tEnable autoindenting\n");
fprintf(stderr,"\t-d     :\tStart '%s' in document mode\n", ourname);
fprintf(stderr,"\t-i     :\tStart '%s' in overstrike mode\n", ourname);
fprintf(stderr,"\t-n     :\tStart '%s' in non-document mode\n", ourname);
fprintf(stderr,"\t-q     :\tProduce this message\n");
fprintf(stderr,"\t-R     :\tDefault search/replace mode is regexp\n");
fprintf(stderr,"\t-S     :\tStart at top of file\n");
fprintf(stderr,"\t-s     :\tStart '%s' with status line disabled\n",ourname);
fprintf(stderr,"\t-W     :\tDefault to wrapping in search/replace\n");
fprintf(stderr,"\t+NN    :\tGo to line number NN\n");
fprintf(stderr,"\t+/pat/ :\tEdit and search for pattern 'pat'\n");
fprintf(stderr,"\t-F     :\tAdd a tagfile to the list of those scanned\n");
fprintf(stderr,"\t-T NN  :\tSet tabwidth\n");
fprintf(stderr,"\t-t tag :\tSearch for 'tag' on startup\n");
fprintf(stderr,"\t-jrr   :\tStart '%s' with ragged right justification\n", ourname);
fprintf(stderr,"\t-jrj   :\tStart '%s' with smooth right justification\n", ourname);
fprintf(stderr,"\t-jcj   :\tStart '%s' with centre justification\n", ourname);
fprintf(stderr,"\t-l NN  :\tSet left margin\n");
fprintf(stderr,"\t-r NN  :\tSet right margin\n");
fprintf(stderr,"\t-w NN  :\tSet screen width\n");
fprintf(stderr,"\t-h NN  :\tSet screen height\n");


	doexit(-1);
}

/* If the string "TABSNN" on first line, use NN as tabwidth */
searchtw()
{
	unsigned char *cp;
	int val;

	cp = mem_buf;
	while (*cp && *cp != '\r' && *cp != '\n' && cp < end_buf)
	{
		if (scpn(&cp, "TABS", &val))
			tabwidth = val;
		else if (scpn(&cp, "ts=", &val))
			tabwidth = val;
		else if (scpn(&cp, "RM", &val))
			rmargin = val - 1;
		else if (scpn(&cp, "LM", &val))
			lmargin = val - 1;
		else if (scpn(&cp, "DOC", 0))
			nondocmode = 0;
		else if (scpn(&cp, "NONDOC", 0))
			nondocmode = 1;
		else if (scpn(&cp, "JRR", 0))
			justifymode = J_RAGGED;
		else if (scpn(&cp, "JRJ", 0))
			justifymode = J_RIGHT;
		else if (scpn(&cp, "JCJ", 0))
			justifymode = J_CENTRE;
		else if (scpn(&cp, "AI", 0))
			autotabs = 1;
		else
			cp++;
	}
}

scpn(pcp, match, pval)
register char **pcp, *match;
int *pval;
{
	register int len;
	register char *cp;

	cp = *pcp;
	len = strlen(match);
	if (strncmp(cp, match, len) || (pval && !isdigit(cp[len])))
		return 0;

	cp += len;
	if (pval)
	{
		*pval = 0;
		while (isdigit(*cp))
		{
			*pval *= 10;
			*pval += *cp++ - '0';
		}
	}
	*pcp = cp;
	return 1;
}

/*
 * Terminal raw mode
 */
struct termios _termios, _oldtermios;
int ttyset = 0;
int curinfd = -1;

void
getwinsize(prows, pcols)
int *prows, *pcols;
{
	struct winsize ws;
	int ret;

	setcurinfd();
	ret = ioctl(curinfd, TIOCGWINSZ, &ws);
	if (ret < 0)
	{
		fprintf(stderr, "getwinsoze(): ioctl() failed: %d\n", errno);
		exit(1);
	}	
	*prows = ws.ws_row;
	*pcols = ws.ws_col;
}

termstart()
{
	setcurinfd();

	if (!ttyset)
	{
		tst(1, tcgetattr(curinfd, &_termios));
		_oldtermios = _termios;
		ttyset = 1;
		_termios.c_iflag &= ~(IXON|IXOFF);
		_termios.c_lflag &= ~ECHO & ~ECHOE & ~ECHONL & ~ICANON;
		_termios.c_cc[VMIN] = 1;	/* Min. no. of chars */
		_termios.c_cc[VINTR] = 0;	/* Ignore SIGINT */
		_termios.c_cc[VQUIT] = 0;	/* Ignore SIGQUIT */
#ifdef VSTART
		_termios.c_cc[VSTART] = 0;
#endif
#ifdef VSTOP
		_termios.c_cc[VSTOP] = 0;
#endif
#ifdef VSUSP
		_termios.c_cc[VSUSP] = 0;
#endif
#ifdef VDSUSP
		_termios.c_cc[VDSUSP] = 0;
#endif
#ifdef VDISCARD
		_termios.c_cc[VDISCARD] = 0;	/* Reclaim ^O */
#endif
		tst(2, tcsetattr(curinfd, TCSADRAIN, &_termios));
	}
}

termend()
{
	if (ttyset)
	{
		tst(3, tcsetattr(curinfd, TCSADRAIN, &_oldtermios));
		ttyset = 0;
	}
}

tst(n, ret)
{
	extern int errno;
	int errnsave;

	if (ret < 0)
	{
		errnsave = errno;
		fprintf(stderr, "termio() failed[%d]: ", n);
		errno = errnsave;
		perror("");
	}
}

setcurinfd()
{
	if (curinfd < 0)
		curinfd = fileno(stdin);
}

doexit(code)
{
	extern int is_xterm;

	termend();
	if (cutFileName)
		Free(cutFileName);
	if (fnames)
		Free(fnames);
	freeDDEPaths();
	freeRollFiles();
	sedterm();
	leditEnd();
	cleanUpTags();
	if (is_xterm)
	{
		char *e = getenv("XTERM_TITLE");
		if (e == 0)
			e = "xterm";
		xterm_title((char *)0, e);
		/* Turn off pointer mode */
		xterm_pointer_mode(0);
		flushop();
	}
	exit(code);
}
